""" Context managers for code generation. """

# Implement a context manager that allows to temporarily define tags globally.
_cache_context_stack = []


class _CacheContext(object):
    def __init__(self, tags, delete=False):
        self.tags = tags
        self.delete = delete

    def __enter__(self):
        _cache_context_stack.append(self.tags)

    def __exit__(self, exc_type, exc_value, traceback):
        _cache_context_stack.pop()

        if self.delete:
            from dune.perftool.generation.cache import delete_cache_items
            delete_cache_items(condition=" and ".join(self.tags))


def cache_context(*tags, **kwargs):
    return _CacheContext(tags, **kwargs)


def get_context_tags():
    result = tuple()
    for items in _cache_context_stack:
        result = result + items
    return result


_global_context_cache = {}


class _GlobalContext(object):
    def __init__(self, **kwargs):
        self.kw = kwargs

    def __enter__(self):
        self.old_kw = {}
        for k, v in self.kw.items():
            # First store existing values of the same keys
            if k in _global_context_cache:
                self.old_kw[k] = v
            # Now replace the value with the new one
            _global_context_cache[k] = v

    def __exit__(self, exc_type, exc_value, traceback):
        # Delete all the entries from this context
        for k in self.kw.keys():
            del _global_context_cache[k]
        # and restore previously overwritten values
        for k, v in self.old_kw.items():
            _global_context_cache[k] = v


def global_context(**kwargs):
    return _GlobalContext(**kwargs)


def get_global_context_value(key):
    return _global_context_cache[key]
