""" Generators for basis evaluations """

from dune.perftool.generation import (cached,
                                      class_member,
                                      domain,
                                      generator_factory,
                                      iname,
                                      include_file,
                                      instruction,
                                      preamble,
                                      symbol,
                                      temporary_variable,
                                      )
from dune.perftool.pdelab.quadrature import (name_quadrature_position,
                                             quadrature_iname,
                                             )
from dune.perftool.pdelab.geometry import (name_dimension,
                                           name_jacobian_inverse_transposed,
                                           )
from dune.perftool.pdelab.localoperator import (lop_template_ansatz_gfs,
                                                lop_template_test_gfs,
                                                )
from dune.perftool.pdelab.driver import FEM_name_mangling


@symbol
def type_localbasis_cache(element):
    return "LocalBasisCacheWithoutReferences<{}>".format(type_gfs(element))


@class_member("operator")
def define_localbasis_cache(element, name):
    include_file("dune/perftool/localbasiscache.hh", filetag="operatorfile")
    t = type_localbasis_cache(element)
    return "{} {};".format(t, name)


@symbol
def name_localbasis_cache(element):
    name = "cache_{}".format(FEM_name_mangling(element))
    define_localbasis_cache(element, name)
    return name


@preamble
def define_lfs_bound(lfs, bound):
    return 'auto {} = {}.size();'.format(bound, lfs)


@symbol
def name_lfs_bound(lfs):
    # LFS might either be given by an UFL element or by a string describig its name
    from ufl import FiniteElementBase
    if isinstance(lfs, FiniteElementBase):
        return name_lfs_bound(name_lfs(lfs))

    bound = '{}_size'.format(lfs)
    define_lfs_bound(lfs, bound)
    return bound


@generator_factory(cache_key_generator=lambda e, **kw: e)
def name_lfs(element, prefix=None):
    # Omitting the prefix is only valid upon a second call, which will
    # result in a cache hit.
    assert prefix

    # Additionally, element is expected to be a ufl finite element
    from ufl import FiniteElementBase
    assert isinstance(element, FiniteElementBase)

    # Recurse into the given element to define all other local function spaces!
    from ufl import MixedElement
    if isinstance(element, MixedElement):
        for i, subelem in enumerate(element.sub_elements()):
            # TODO in this case, we need to trigger the extraction mechanism
            # as preambles in our code.
            name_lfs(subelem, prefix + "_" + str(i))

    # Now trigger the creation of all those other symbols/preables necessary for this lfs

    # Now return the prefix!
    return prefix


@generator_factory(cache_key_generator=lambda e, **kw: e)
def type_gfs(element, basetype=None, index_stack=None):
    # Omitting basetype and index_stack is only valid upon a second call,
    # which will result in a cache hit.
    assert basetype
    assert index_stack is not None

    # Additionally, element is expected to be a ufl finite element
    from ufl import FiniteElementBase
    assert isinstance(element, FiniteElementBase)

    # Recurse into the given element to define all other local function spaces!
    from ufl import MixedElement
    if isinstance(element, MixedElement):
        for i, subelem in enumerate(element.sub_elements()):
            type_gfs(subelem, basetype=basetype, index_stack=index_stack + (i,))

    if len(index_stack) == 0:
        return basetype
    else:
        include_file("dune/typetree/childextraction.hh", filetag="operatorfile")
        return 'Dune::TypeTree::Child<{},{}>'.format(basetype, ','.join(str(i) for i in index_stack))


def traverse_lfs_tree(arg):
    from dune.perftool.ufl.modified_terminals import ModifiedArgumentDescriptor
    assert isinstance(arg, ModifiedArgumentDescriptor)

    # First we need to determine the basename as given in the signature of
    # this kernel method!
    lfs_basename = None
    gfs_basename = None
    from ufl.classes import Argument, Coefficient
    if isinstance(arg.argexpr, Argument):
        if arg.argexpr.count() == 0:
            lfs_basename = 'lfsv'
            gfs_basename = 'GFSV'
        if arg.argexpr.count() == 1:
            lfs_basename = 'lfsu'
            gfs_basename = 'GFSU'
        # TODO add restrictions here.

    if isinstance(arg.argexpr, Coefficient):
        # We should only ever call this for a trialfunction, which in our case
        # is the coefficient of reserved index 0.
        assert arg.argexpr.count() == 0

        lfs_basename = 'lfsu'
        gfs_basename = 'GFSU'

    assert lfs_basename and gfs_basename

    # Now start recursively extracting local function spaces and fill the cache with
    # all those values. That way we can later get a correct local function space with
    # just the ufl finite element.
    name_lfs(arg.argexpr.element(), prefix=lfs_basename)
    type_gfs(arg.argexpr.element(), basetype=gfs_basename, index_stack=())


@iname
def _lfs_iname(element, argcount, context):
    name = name_lfs(element)
    bound = name_lfs_bound(name)

    if argcount != 0:
        name = 'lfsu'

    if context:
        context = '_' + context

    name = name + context + '_index'
    domain(name, bound)

    return name


def lfs_iname(element, argcount=0, context=''):
    """ Get the iname to iterate over the local function space given by element

    Arguments:
    ----------
    element: ufl.FiniteElementBase
        The finite element this local function space belongs to
    argcount: int
        Use to realize double nesting in case of jacobians
    context: str
        Some generation methods will require you to duplicate an iname for
        a given purpose, see the 'Loops and dependencies' of the loopy docs:
        https://documen.tician.de/loopy/tutorial.html#loops-and-dependencies
    """
    return _lfs_iname(element, argcount, context)


@preamble
def declare_cache_temporary(element, name, which):
    t_cache = type_localbasis_cache(element)
    lfs = name_lfs(element)
    return "typename {}::{}ReturnType {};".format(t_cache,
                                                  which,
                                                  name,
                                                  )


@cached
def evaluate_basis(element, name):
    temporary_variable(name, shape=(name_lfs_bound(element),), decl_method=None)
    declare_cache_temporary(element, name, which='Function')
    cache = name_localbasis_cache(element)
    lfs = name_lfs(element)
    qp = name_quadrature_position()
    instruction(inames=(quadrature_iname(),
                        ),
                code='{} = {}.evaluateFunction({}, {}.finiteElement().localBasis());'.format(name,
                                                                                             cache,
                                                                                             qp,
                                                                                             lfs,
                                                                                             ),
                assignees=frozenset({name}),
                )


@symbol
def name_basis(element):
    evaluate_basis(element, "phi")
    return "phi"


@cached
def evaluate_reference_gradient(element, name):
    temporary_variable(name, shape=(name_lfs_bound(element), name_dimension()), decl_method=None)
    declare_cache_temporary(element, name, which='Jacobian')
    cache = name_localbasis_cache(element)
    lfs = name_lfs(element)
    qp = name_quadrature_position()
    instruction(inames=(quadrature_iname(),
                        ),
                code='{} = {}.evaluateJacobian({}, {}.finiteElement().localBasis());'.format(name,
                                                                                             cache,
                                                                                             qp,
                                                                                             lfs,
                                                                                             ),
                assignees=frozenset({name}),
                )


@symbol
def name_reference_gradient(element):
    evaluate_reference_gradient(element, "js")
    return "js"


@cached
def evaluate_basis_gradient(element, name):
    # TODO this is of course not yet correct
    temporary_variable(name, shape=(name_lfs_bound(element), name_dimension()), shape_impl=('vec', 'fv'))
    jac = name_jacobian_inverse_transposed()
    index = lfs_iname(element, context='transformgrads')
    reference_gradients = name_reference_gradient(element)
    instruction(inames=(index,
                        quadrature_iname(),
                        ),
                code='{}.mv({}[{}][0], {}[{}]);'.format(jac,
                                                        reference_gradients,
                                                        index,
                                                        name,
                                                        index,
                                                        ),
                assignees=name,
                read_variables=frozenset({reference_gradients}),
                )


@symbol
def name_basis_gradient(element):
    evaluate_basis_gradient(element, "gradphi")
    return "gradphi"


@cached
def evaluate_trialfunction(element, name):
    temporary_variable(name, shape=())
    lfs = name_lfs(element)
    index = lfs_iname(element)
    basis = name_basis()
    instruction(inames=(quadrature_iname(),
                        index,
                        ),
                code='{} += x({}, {})*{}[{}];'.format(name,
                                                      lfs,
                                                      index,
                                                      basis,
                                                      index
                                                      ),
                assignees=frozenset({name}),
                read_variables=frozenset({basis}),
                )


def reset_trialfunction_gradient(name):
    return instruction(inames=(quadrature_iname(),
                               ),
                       code='{} = 0.0;'.format(name),
                       assignees=frozenset({name}),
                       )


@cached
def evaluate_trialfunction_gradient(element, name):
    # TODO this is of course not yet correct
    temporary_variable(name, shape=(name_dimension(),), shape_impl=('fv',))
    reset = reset_trialfunction_gradient(name)
    lfs = name_lfs(element)
    index = lfs_iname(element, context='trialgrad')
    basis = name_basis_gradient(element)
    instruction(inames=(quadrature_iname(),
                        index,
                        ),
                code='{}.axpy(x({}, {}), {}[{}]);'.format(name,
                                                          lfs,
                                                          index,
                                                          basis,
                                                          index,
                                                          ),
                assignees=frozenset({name}),
                read_variables=frozenset({basis}),
                forced_iname_deps=frozenset({quadrature_iname(), index}),
                forced_iname_deps_is_final=True,
                depends_on=frozenset({reset}),
                )
